/*
 * PlasmWrapper.java
 *
 * Created on 29 settembre 2007, 8.35
 *
 * 
 * @author Dario Lucia
 *
 */

package it.jplasmide.utils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class PlasmWrapper {
    
    public static final String COMMAND_IN = "__!-- COMMAND IN --!__";
    public static final String COMMAND_SENT = "__!-- COMMAND SENT --!__";
    
    private final String ESCAPE_SEQUENCE = "#??_!!_0-0_!!_??#";
    private final String MZSCHEME_VAR = "MZSCHEME";
    
    private String NEW_LINE = System.getProperty("line.separator");
    private Process plasmProcess = null;
    
    private PlasmReaderThread reader;
    private PlasmWriterThread writer;
    
    private List<PlasmEventListener> listeners;

    private boolean nokill = false;

    private Lock plasmStopLock = new ReentrantLock();
    
    /** Creates a new instance of PlasmWrapper */
    public PlasmWrapper() {
       this.listeners = new ArrayList<PlasmEventListener>();
    }
    
    /** Start the plasm interpreter */
    public void startPlasm(String plasmPath, String mzschemePath) {
   
        if(this.plasmProcess == null) {
            
            /* Get all env var*/
            Runtime rt = Runtime.getRuntime();
            
            Map<String,String> envs = System.getenv();
            List<String> envList = new LinkedList<String>();
            
            for(Map.Entry<String,String> me : envs.entrySet()) {
                if(!me.getKey().equals(MZSCHEME_VAR)) {
                    envList.add(me.getKey() + "=" + me.getValue());
                }
            }
             
            envList.add(MZSCHEME_VAR + "=" + mzschemePath);
             
//            String display = System.getenv("DISPLAY");
//            
//            System.out.println("DISPLAY : " + display);
//            
            String[] cmds = { plasmPath };
            String[] env = envList.toArray(new String[0]);
             
//            String[] env = { MZSCHEME_VAR + "=" +  mzschemePath ,
//            "DISPLAY=" + display ,    "HOME=" +  System.getProperty("user.home")  };

            try {
                System.out.println("Command: " + cmds[0]);
                String currDir = plasmPath.substring(0,plasmPath.lastIndexOf(System.getProperty("file.separator")));
                this.plasmProcess = rt.exec(cmds, env,new File(currDir));
                
                notifyPlasmStart();
                
                this.reader = new PlasmReaderThread();
                this.writer = new PlasmWriterThread();
                
                /* Starting threads */
                this.writer.plasmStarted(this);
                this.reader.plasmStarted(this);
               
            
            } catch (IOException ex) {
                ex.printStackTrace();
                notifyPlasmError(ex.getMessage());
                notifyPlasmStop();
                
                this.plasmProcess = null;
                
            }
            
        } else {
            notifyPlasmError("Plasm process already running" + NEW_LINE);
        }
    }
    
     /** Stop the plasm interpreter */
    public void stopPlasm() {
        
        try {
            if(plasmProcess != null) {
                /* Stop the writer */
               this.writer.sendCommand(ESCAPE_SEQUENCE);
                /* Signal */
               notifyPlasmStop();

               /* Cleanup */
               this.reader = null;
               this.writer = null;

               try {
                this.plasmProcess.destroy();
               } catch (Exception e) {
                   e.printStackTrace();
               }

               this.plasmProcess = null;
            } else {
                notifyPlasmError("No process to stop" + NEW_LINE);
            }
        } catch (Exception e) {
            System.out.println("Exception during stopPlasm(): " + e.getMessage());
        }
        
    }
    
    public void plasmStopped() {
        System.out.println("Invoking plasmStopped()");
       
        if(!this.nokill) {
            try {
           /* Stop the writer */
           this.writer.sendCommand(ESCAPE_SEQUENCE);
            } catch(NullPointerException e) {
                System.out.println("Catch null exception.. Dont worry, it's normal (damn concurrency ... >_>)");
            }
            
            /* Signal */
           notifyPlasmStop();

           /* Cleanup */
           this.reader = null;
           this.writer = null;
           
           try {
               if(this.plasmProcess != null)
                    this.plasmProcess.destroy();
           } catch (Exception e) {
               e.printStackTrace();
           }
           
           this.plasmProcess = null;
       } else {
            notifyPlasmError("Lost communication with plasm process" + NEW_LINE);
            notifyPlasmError("Maybe it would be better to restart PLaSM interpreter" + NEW_LINE);
       }
      
    }
    
    public void evalPlasmCommand(String t) {
        if(this.writer != null) {
            StringBuffer b = new StringBuffer();
            b.append("(plasm \" ");
            b.append(NEW_LINE);
            b.append(t);
            b.append(NEW_LINE);
            b.append("\") ");

            this.writer.sendCommand(b.toString());
        } else {
            notifyPlasmError("Cannot send Plasm commands to interpreter - Writer error" + NEW_LINE);
        }
    }
    
    public void evalSchemeCommand(String t) {
        if(this.writer != null)
            this.writer.sendCommand(t);
        else 
            notifyPlasmError("Cannot send Scheme commands to interpreter - Writer error" + NEW_LINE);
    
    }
    
    public void addPlasmListener(PlasmEventListener e) {
        this.listeners.add(e);
    }
    
    public void removePlasmListener(PlasmEventListener e) {
        this.listeners.remove(e);
    }

    void notifyPlasmError(String string) {
        for(PlasmEventListener l : this.listeners)
            l.plasmErrorEvent(string);
    }
    
    void notifyPlasmStop() {
        for(PlasmEventListener l : this.listeners)
            l.plasmStoppedEvent();
    }
    
    void notifyPlasmStart() {
        for(PlasmEventListener l : this.listeners)
            l.plasmStartedEvent();
    }
    
    void notifyPlasmCatchedOutputEvent(String string) {
        for(PlasmEventListener l : this.listeners)
            l.plasmCatchedOutputEvent(string);
    }
    
    void notifyPlasmRequestInputEvent(String string) {
        for(PlasmEventListener l : this.listeners)
            l.plasmRequestInputEvent(string);
    }
    
    Process getPlasmCurrentProcess() {
        return this.plasmProcess;
    }

    String getEscapeSequence() {
        return ESCAPE_SEQUENCE;
    }

    public void setNoKillOnBrokenPipe(boolean b) {
        this.nokill = b;
    }
    
}
